/*******************************************************************
 * File: gui.cpp (Win32 Version)
 * Author: Keith Schwarz (htiek@cs.stanford.edu)
 *
 * Windows-specific implementation of the GUI framework and user
 * interaction code.  You should not need to modify the contents of
 * this file.  Because this is Windows-specific, most of the code here
 * is not part of the C++ standard and you are not required to
 * understand how it works.  However, if you'd like a quick intro to
 * Win32 programming in C++, send me an email and I'd love to give you
 * a tutorial.
 */

/* Windows-specific headers.  windows.h contains most of the definitions
 * for Windows-specific functions.  The tchar.h header defines the TCHAR
 * type, which is similar to char except that it supports unicode and
 * multibyte character formats.
 */
#include <windows.h> // For Win32 API
#include <tchar.h>   // For _T macro

/* Resource definitions.  Windows supports the idea of "resources;"
 * data embedded into an executable file containing useful information.
 * In this particular case, the resource file contains the layout of
 * the dialog boxes.  This header allows us to reference the resources
 * in this program.
 */
#include "resource"

/* Standard library includes. */
#include <string>
#include <sstream>
#include <cassert>
#include <cstdlib>
#include <vector>
#include <fstream>

/* Email-specific includes. */
#include "EmailMessage.h"
#include "mail.h"
using namespace std;

/* Prototypes for implementation functions. */
static void Reset(HWND hDlg);
static BOOL CALLBACK AddressBookDialogProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);

/* Code that the user sent correctly/incorrectly.  These values are arbitrary
 * but must be nonzero.
 */
const int kSuccessCode = 137;
const int kFailCode    = kSuccessCode + 1;

/* Globals referenced by the GUI hooks.  I know that global variables are an
 * abomination, but for this particular application the complexity of using
 * only local variables is hardly justified.
 */
static EmailMessage   gMessage;
static vector<string> gAddresses;

/* TCHAR string type.  Windows applications use a special type called TCHAR
 * for all characters that allows the code to compile both on ASCII machines
 * and on multibyte character set machines (i.e. Unicode).  This code builds
 * a string type called tstring using the standard library's basic_string
 * template.  Interestingly, basic_string is what's used to build the standard
 * string type; string is actually a typedef for basic_string<char>.
 */
typedef basic_string<TCHAR> tstring;

/* Conversion functions to convert between C++ strings and TCHAR strings. */
static tstring WidenString(string str)
{
	tstring result;
	for (size_t i = 0; i < str.length(); ++i)
		result += str[i];
	return result;
}
static string ConstrictString(tstring str)
{
	string result;
	for (size_t i = 0; i < str.length(); ++i)
		result += char(str[i]);
	return result;
}

/* Strings in the address window are described using the format
 *
 * To: htiek@cs.stanford.edu or
 * Cc: htiek@cs.stanford.edu
 *
 * These functions parse these messages apart to yield the recipient
 * and what sort of recipient they are (i.e. To, Cc, Bcc)
 */
static EmailMessage::RecipientType ExtractRecipientType(tstring msg)
{
	if (msg.find(_T("To: ")) == 0)
		return EmailMessage::To;
	if (msg.find(_T("Cc: ")) == 0)
		return EmailMessage::Cc;
	if (msg.find(_T("Bcc: ")) == 0)
		return EmailMessage::Bcc;

	assert (0 && "Invalid recipient?");
	return EmailMessage::To; // Suppress warnings.
}

static tstring ExtractRecipient(tstring msg)
{
	/* CHeck that we can find a separator. */
	assert (msg.find(_T(": ")) != string::npos);

	/* Skip over the type and colon. */
	msg.erase(0, msg.find(_T(": ")) + 2);
	return msg;
}

/* All lists and text boxes in Win32 are actually windows that are
 * nested inside of other windows.  The code for extracting information
 * from these windows is fairly dense, and to simplify the code for
 * extracting this data I've decomposed out a few useful functions to
 * get the data.  This function reads a string out of a list box, a
 * window with a list of strings in it.
 */
tstring StringFromList(HWND list, int index)
{
	tstring str;
	str.resize(SendMessage(list, LB_GETTEXTLEN, index, 0) + 1);
	SendMessage(list, LB_GETTEXT, index, reinterpret_cast<LPARAM>(&str[0]));
	return str.substr(0, str.length() - 1); // Drop null terminator
}

/* An edit control is a text box.  This code reads the contents of that text box
 * into a tstring.
 */
tstring StringFromEditControl(HWND list)
{
	tstring str;
	str.resize(SendMessage(list, WM_GETTEXTLENGTH, 0, 0) + 1);
	SendMessage(list, WM_GETTEXT, str.length(), reinterpret_cast<LPARAM>(&str[0]));
	return str.substr(0, str.length() - 1); // Drop null terminator.
}

/* Populates the address book with data provided by your LoadAddressBook() implementation. */
static void PopulateAddressBook(HWND addressBook, HWND recipients)
{
	gAddresses = LoadAddressBook(); // Reload the address book.

	/* Reset the address book. */
	SendMessage(addressBook, LB_RESETCONTENT, 0, 0);
	for (size_t i = 0; i < gAddresses.size(); ++i)
		SendMessage(addressBook, LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(WidenString(gAddresses[i]).c_str()));

	/* Also reset the recipients. */
	SendMessage(recipients, LB_RESETCONTENT, 0, 0);
}

/* After the user clicks send, the data from the window must be converted into
 * an EmailMessage object.  This function extracts the data from the window and
 * fills in the global EmailMessage object with it.
 */
static void SaveInfo(HWND hDlg)
{
	gMessage.clear();

	HWND subject     = GetDlgItem(hDlg, IDC_SUBJECTTEXT);
	HWND message     = GetDlgItem(hDlg, IDC_MESSAGETEXT);
	HWND recipients  = GetDlgItem(hDlg, IDC_RECIPIENTLIST);
	HWND attachments = GetDlgItem(hDlg, IDC_ATTACHMENTSLIST);

	assert (message && subject && recipients && attachments);

	/* Set primary fields. */
	gMessage.setSender(UserName());
	gMessage.setSubject(ConstrictString(StringFromEditControl(subject)));
	gMessage.setMessageText(ConstrictString(StringFromEditControl(message)));

	/* Add recipients. */
	const int numRecipients = int(SendMessage(recipients, LB_GETCOUNT, 0, 0));
	for (int i = 0; i < numRecipients; ++i)
	{
		tstring recipient = StringFromList(recipients, i);
		gMessage.addRecipient(ConstrictString(ExtractRecipient(recipient)),
			                  ExtractRecipientType(recipient));
	}

	/* Add attachments. */
	const int numAttachments = int(SendMessage(attachments, LB_GETCOUNT, 0, 0));
	for (int i = 0; i < numAttachments; ++i)
		gMessage.addAttachment(ConstrictString(StringFromList(attachments, i)));
}

/* After the user clicks Send, we need to make sure that the user has entered
 * enough information to form a valid email.  This function checks that the
 * provided information is sufficient.
 */
static bool IsValidInfo(HWND hDlg)
{
	HWND recipients = GetDlgItem(hDlg, IDC_RECIPIENTLIST);

	assert (recipients);

	/* Check that there's at least one recipient. */
	const LRESULT numRecipients = SendMessage(recipients, LB_GETCOUNT, 0, 0);
	if (numRecipients == LB_ERR || numRecipients == 0)
	{
		MessageBox(NULL, _T("You need to specify one or more recipients for this message."),
						 _T("Send"), MB_OK | MB_ICONEXCLAMATION);
		return false;
	}

	return true;
}

/* Resets the compose dialog. */
static void Reset(HWND hDlg)
{	
	gMessage.clear();    // Clear the email message.
	
	HWND addressBook = GetDlgItem(hDlg, IDC_ADDRESSBOOKLIST);
	HWND recipients  = GetDlgItem(hDlg, IDC_RECIPIENTLIST);
	HWND subject     = GetDlgItem(hDlg, IDC_SUBJECTTEXT);
	HWND body        = GetDlgItem(hDlg, IDC_MESSAGETEXT);
	HWND attachments = GetDlgItem(hDlg, IDC_ATTACHMENTSLIST);
	assert (addressBook && recipients && subject && body && attachments);

	PopulateAddressBook(addressBook, recipients);

	/* ... and the attachments. */
	SendMessage(attachments, LB_RESETCONTENT, 0, 0);

	/* Clear the subject and body. */
	SetWindowText(subject, _T(""));
	SetWindowText(body,    _T(""));
}

/* Removes a recipient from the list and puts her back into the address book. */
static void RemoveRecipient(HWND hDlg)
{
	assert (hDlg);

	HWND addressBook = GetDlgItem(hDlg, IDC_ADDRESSBOOKLIST);
	HWND recipients  = GetDlgItem(hDlg, IDC_RECIPIENTLIST);

	assert (addressBook && recipients);

	/* Make sure something's selected. */
	LRESULT index = SendMessage(recipients, LB_GETCURSEL, 0, 0);
	if (index == LB_ERR) return;

	/* Retrieve the string.  This requires us to allocate a copy of the string
	 * on our end first.
	 */
	tstring str = StringFromList(recipients, int(index));

	/* Remove the string from the source. */
	SendMessage(recipients, LB_DELETESTRING, index, 0);

	/* Crop the message label. */
	str = ExtractRecipient(str);

	/* Add it to the destination with the appropriate prefix. */
	SendMessage(addressBook, LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(str.c_str()));
}

/* Given a type of recipient, adds the selected user from the user list to
 * the recipient list.
 */
static void AddRecipient(HWND hDlg, EmailMessage::RecipientType type)
{
	assert (hDlg);

	HWND addressBook = GetDlgItem(hDlg, IDC_ADDRESSBOOKLIST);
	HWND recipients  = GetDlgItem(hDlg, IDC_RECIPIENTLIST);

	assert (addressBook && recipients);

	/* Make sure something is selected! */
	LRESULT index = SendMessage(addressBook, LB_GETCURSEL, 0, 0);
	if (index == LB_ERR) return;

	/* Retrieve the string.  This requires us to allocate a copy of the string
	 * on our end first.
	 */
	tstring str = StringFromList(addressBook, int(index));

	/* Remove the string from the source. */
	SendMessage(addressBook, LB_DELETESTRING, index, 0);

	/* Based on the type of the message, prepend some string to the name. */
	switch (type)
	{
	case EmailMessage::To:
		str = _T("To: ") + str;
		break;
	case EmailMessage::Cc:
		str = _T("Cc: ") + str;
		break;
	case EmailMessage::Bcc:
		str = _T("Bcc: ") + str;
		break;
	default:
		/* Should be unreachable. */
		assert(0 && "How did this happen?");
	}

	/* Add it to the destination with the appropriate prefix. */
	SendMessage(recipients, LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(str.c_str()));
}

/* Lets the user pick a file to attach.  This uses one of the most complicated functions
 * I've ever seen - GetOpenFileName - to prompt the user for a file.  We then add it to
 * the attachments list.
 */
static void AttachFile(HWND hDlg)
{
	/* Save the current directory... otherwise getting the open file will change
	 * it and we'll be in the wrong place!
	 */
	TCHAR currDirectory[MAX_PATH + 1];
	GetCurrentDirectory(MAX_PATH, currDirectory);

	/* Holds where the filename will be stored. */
	TCHAR resultBuffer[MAX_PATH + 1] = {0};

	OPENFILENAME ofn;
	memset(&ofn, 0, sizeof(ofn)); // Zero the object.

	ofn.lStructSize			= sizeof(OPENFILENAME);
	ofn.hwndOwner			= hDlg;
	ofn.hInstance			= NULL;
	ofn.lpstrFilter			= _T("All Files\0*.*\0\0");
	ofn.lpstrCustomFilter	= NULL;
	ofn.nFilterIndex        = 1;
	ofn.lpstrFile           = resultBuffer;
	ofn.nMaxFile            = MAX_PATH;
	ofn.lpstrFileTitle      = NULL;
	ofn.lpstrInitialDir     = NULL;
	ofn.lpstrTitle          = _T("Choose File to Attach");
	ofn.Flags               = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;

	/* Try to get an open file.  If we do, add it to the attachments list. */
	if (GetOpenFileName(&ofn))
		SendMessage(GetDlgItem(hDlg, IDC_ATTACHMENTSLIST), LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(resultBuffer));

	/* Restore the directory. */
	SetCurrentDirectory(currDirectory);
}

/* Brings up the "edit address book" dialog and works with it. */
static void ManageAddressBook(HWND hDlg)
{
	/* Launch the dialog box. */
	LRESULT result = DialogBoxParam(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_EDITADDRESSES), hDlg, AddressBookDialogProc, 0);

	/* If the user hit "OK," repopulate the address book with the updated list. */
	if (result == kSuccessCode)
	{	
		/* Save the addresses we entered. */
		SaveAddressBook(gAddresses);

		/* Update the address book list. */
		HWND addressBook = GetDlgItem(hDlg, IDC_ADDRESSBOOKLIST);
		HWND recipients  = GetDlgItem(hDlg, IDC_RECIPIENTLIST);

		assert (addressBook && recipients);
		PopulateAddressBook(addressBook, recipients);
	}
}

/* Windowed programs are event-driven, meaning that the operating system will call a specific
 * function registered by the program whenever something happens.  This function is the callback
 * for the Compose dialog box.  Understanding how this works requires a fairly good understanding of
 * the Win32 API, so if you're interested in seeing how everything works together come talk to me
 * after lecture or send me an email.
 */
static BOOL CALLBACK MainDialogProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM)
{
	/* Figure out what message we received. */
	switch(message)
	{
	/* When the window is created, initialize all of the fields appropriately. */
	case WM_INITDIALOG:
		Reset(hDlg);
		return TRUE;

	/* React to child controls. */
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		/* Send messages. */
		case IDC_SEND:
			/* Check whether the info is valid, and send if it is. */
			if (IsValidInfo(hDlg))
			{
				/* Save the information into the message. */
				SaveInfo(hDlg);

				/* Call your SendEmail function to send the email and display the result! */
				if (SendEmail(gMessage))
				{
					MessageBox(hDlg, _T("Message sent successfully!"), _T("Send"), MB_OK | MB_ICONINFORMATION);
					Reset(hDlg);
				}
				else
					MessageBox(hDlg, _T("Message send failed."), _T("Send"), MB_OK | MB_ICONSTOP);
			}
			break;

		/* User hits the "Clear" button. */
		case IDC_RESET:
			Reset(hDlg);
			break;

		/* Add recipients. */
		case IDC_ADDTO:
			AddRecipient(hDlg, EmailMessage::To);
			break;
		case IDC_ADDCC:
			AddRecipient(hDlg, EmailMessage::Cc);
			break;
		case IDC_ADDBCC:
			AddRecipient(hDlg, EmailMessage::Bcc);
			break;

		/* User removes a recipent. */
		case IDC_REMOVERECIPIENT:
			RemoveRecipient(hDlg);
			break;

		/* User clicks the attach button. */
		case IDC_ATTACHBUTTON:
			AttachFile(hDlg);
			break;

		/* User clicks the "Manage Addresses" button. */
		case IDC_MANAGEADDRESSESBUTTON:
			ManageAddressBook(hDlg);
			break;
		}
		break;

	/* User hits the 'close' button. */
	case WM_CLOSE:
		EndDialog(hDlg, kFailCode);
		break;
	}

	return FALSE;
}

/* Populates the address book with all of the user's addresses. */
static void InitializeAddressBook(HWND hDlg)
{
	assert (hDlg);

	HWND addressBook = GetDlgItem(hDlg, IDC_ADDRESSBOOKLIST);

	assert (addressBook);

	for (size_t i = 0; i < gAddresses.size(); ++i)
		SendMessage(addressBook, LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(WidenString(gAddresses[i]).c_str()));
}

/* Determines whether a string is a valid email address.  This function is based on a
 * beautiful idea from theoretical computer science called a deterministic finite
 * automaton, or DFA.  Unfortunately, this code is not at all easy to read because it's
 * representing a DFA as efficiently as possible.  Come talk to me after lecture or
 * send me an email for more information about how this works.
 */
static bool IsValidEmailAddress(string str)
{
	/* Search for  a+ [. a+]* @ a* [. a+]+
	 *           0 1   2  3   4 5   6 7
	 */
	int state = 0;
	for (size_t i = 0; i < str.length(); ++i)
	{
		/* Can be alnum, ., or @. */
		if (isalnum(str[i]) || str[i] == '-')
		{
			switch(state)
			{
			case 0: state = 1; break;
			case 1: break;
			case 2: state = 3; break;
			case 3: break;
			case 4: state = 5; break;
			case 5: break;
			case 6: state = 7; break;
			case 7: break;
			}
		}
		else if (str[i] == '.')
		{
			switch(state)
			{
			case 0: return false;
			case 1: state = 2; break;
			case 2: return false;
			case 3: state = 2; break;
			case 4: return false;
			case 5: state = 6; break;
			case 6: return false;
			case 7: state = 6; break;
			}
		}
		else if (str[i] == '@')
		{
			switch(state)
			{
			case 0: return false;
			case 1: state = 4; break;
			case 2: return false;
			case 3: state = 4; break;
			case 4: return false;
			case 5: return false;
			case 6: return false;
			case 7: return false;
			}
		}
		else return false;
	}
	return state == 7;
}

/* In the Manage Addresses dialog, adds a new address to the address list. */
static void AddNewAddress(HWND hDlg)
{
	/* Get the text from the user and see that it's a valid email address. */
	tstring address = StringFromEditControl(GetDlgItem(hDlg, IDC_NEWADDRESS));
	if (!IsValidEmailAddress(ConstrictString(address)))
	{
		MessageBox(hDlg, _T("Please enter a valid email address."), _T("Add Address"), MB_OK | MB_ICONINFORMATION);
		return;
	}

	/* Check whether it's already in the list. */
	HWND addressBook = GetDlgItem(hDlg, IDC_ADDRESSBOOKLIST);
	assert (addressBook);

	const int numAddresses = int(SendMessage(addressBook, LB_GETCOUNT, 0, 0));
	for (int i = 0; i < numAddresses; ++i)
	{
		if (StringFromList(addressBook, i) == address)
		{
			MessageBox(hDlg, _T("That email address is already in your address book."),
							 _T("Add Address"), MB_OK | MB_ICONINFORMATION);
			return;
		}
	}

	/* Add it! */
	SendMessage(addressBook, LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(&address[0]));
	SetWindowText(GetDlgItem(hDlg, IDC_NEWADDRESS), L"");
}

/* Remove the currently-selected email address from the list. */
static void DeleteAddress(HWND hDlg)
{
	assert (hDlg);

	HWND addressBook = GetDlgItem(hDlg, IDC_ADDRESSBOOKLIST);

	assert (addressBook);

	/* Make sure something is selected! */
	LRESULT index = SendMessage(addressBook, LB_GETCURSEL, 0, 0);
	if (index == LB_ERR) return;

	/* Remove the string from the source. */
	SendMessage(addressBook, LB_DELETESTRING, index, 0);
}

/* Saves the list of addresses generated by the user. */
static void UpdateAddressBook(HWND hDlg)
{
	assert (hDlg);

	HWND addressBook = GetDlgItem(hDlg, IDC_ADDRESSBOOKLIST);
	assert (addressBook);

	/* Populate the address book with this data. */
	gAddresses.clear();

	const int numAddresses = int(SendMessage(addressBook, LB_GETCOUNT, 0, 0));
	for (int i = 0; i < numAddresses; ++i)
		gAddresses.push_back(ConstrictString(StringFromList(addressBook, i)));

	SaveAddressBook(gAddresses);
}

/* Event processor for the "manage address book" popup. */
static BOOL CALLBACK AddressBookDialogProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM)
{
	switch(message)
	{
	case WM_INITDIALOG:
		InitializeAddressBook(hDlg);
		return TRUE;

	/* React to button presses */
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		/* Add address. */
		case IDC_ADDADDRESS:
			AddNewAddress(hDlg);
			break;

		/* Remove address. */
		case IDC_DELETEADDRESS:
			DeleteAddress(hDlg);
			break;

		/* Accept and cancel buttons. */
		case IDC_OK:
			UpdateAddressBook(hDlg);
			EndDialog(hDlg, kSuccessCode);
			break;
		case IDC_CANCEL:
			EndDialog(hDlg, kFailCode);
			break;
		}
		break;

	/* User clicks the 'close' button. */
	case WM_CLOSE:
		EndDialog(hDlg, kFailCode);
		break;
	}

	return FALSE;
}

/* Win32 applications begin execution in a special function called WinMain.  Our
 * particular WinMain function launches the Compose dialog box and quits when the
 * dialog box closes.
 */
int WINAPI WinMain(HINSTANCE, HINSTANCE, LPSTR, int)
{
	DialogBoxParam(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_COMPOSE), NULL, MainDialogProc, 0);
	return 0;
}
